/*----------------------------------------------------------------------
  File    : som.c
  Contents: simple self-organizing map management
  Author  : Christian Borgelt
  History : 2000.06.18 file created from file xsom.c
            2001.07.07 extended to 3-dim. reference vectors
            2001.07.08 function som_draw added
            2001.12.21 cube region added
            2002.10.22 METAPOST output function added
            2013.08.09 adapted to higher compiler warning level
----------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <float.h>
#include <math.h>
#include <assert.h>
#include "somd.h"

#ifndef INFINITY
#define INFINITY     (DBL_MAX+DBL_MAX)
#endif                          /* in case C99 is not supported */

/*----------------------------------------------------------------------
  Preprocessor Definitions
----------------------------------------------------------------------*/
#ifndef M_PI                    /* if not defined (no Posix source) */
#define M_PI      3.14159265358979323846  /* \pi */
#define M_SQRT2   1.41421356237309504880
#endif

/*----------------------------------------------------------------------
  Self-Organizing Map Functions
----------------------------------------------------------------------*/

SOM* som_create (int width, int height)
{                               /* --- create a self-organizing map */
  SOM    *som;                  /* created self-organizing map */
  REFVEC *v, **p;               /* to traverse the reference vectors */

  assert((width > 0) && (height > 0));
  som = (SOM*)malloc(sizeof(SOM) +(size_t)(height-1)*sizeof(REFVEC*));
  if (!som) return NULL;        /* create the map body */
  som->width  = width;          /* and initialize it, then */
  som->height = height;         /* create the reference vectors */
  v = (REFVEC*)malloc((size_t)width *(size_t)height *sizeof(REFVEC));
  if (!v) { free(som); return NULL; }
  for (p = som->rvecs; --height >= 0; ) {
    *p++ = v; v += width; }     /* organize the reference vectors */
  return som;                   /* return the created self-org. map */
}  /* som_create() */

/*--------------------------------------------------------------------*/

void som_delete (SOM *som)
{                               /* --- delete a self-organizing map */
  assert(som);                  /* check the function argument */
  free(som->rvecs[0]);          /* delete the reference vectors */
  free(som);                    /* and the map body */
}  /* som_delete() */

/*--------------------------------------------------------------------*/

void som_init (SOM *som, double randfn(void),
               double x, double y, double z)
{                               /* --- init. a self-organizing map */
  int    ix, iy;                /* loop variables */
  REFVEC *rv;                   /* to traverse the reference vectors */

  assert(som && randfn);        /* check the function arguments */
  for (iy = som->height; --iy >= 0; ) {
    for (rv = som->rvecs[iy] +(ix = som->width); --ix >= 0; ) {
      (--rv)->x = 2*x *randfn() -x;
         rv ->y = 2*y *randfn() -y;
         rv ->z = 2*z *randfn() -z;
    }                           /* traverse the reference vectors and */
  }                             /* set random weights in the interval */
}  /* som_init() */             /* [-range, +range] */

/*--------------------------------------------------------------------*/

void som_train (SOM *som, double x, double y, double z,
                double lrate, int nbhood, double radius)
{                               /* --- train a self-organizing map */
  int    ix, iy;                /* loop variables, indices */
  int    bx, by;                /* indices of best reference vector */
  int    xmin, xmax;            /* horizontal range of units */
  int    ymin, ymax;            /* vertical   range of units */
  REFVEC *rv;                   /* to traverse the reference vectors */
  double dx, dy, dz;            /* distances to reference vectors */
  double d, min;                /* (minimal) squared distance */
  int    r, r2;                 /* integer radius and its square */
  int    di, dxi, dyi;          /* integer distances */

  assert(som);                  /* check the function argument */
  min = INFINITY; bx = by = 0;  /* traverse the reference vector */
  for (iy = som->height; --iy >= 0; ) {
    for (rv = som->rvecs[iy] +(ix = som->width); --ix >= 0; ) {
      dx = (--rv)->x -x;        /* for each reference vector */
      dy =    rv ->y -y;        /* compute the squared distance */
      dz =    rv ->z -z;        /* to the training pattern */
      d  = dx*dx + dy*dy + dz*dz;
      if (d < min) { min = d; bx = ix; by = iy; }
    }                           /* find the "winner neuron", i.e., */
  }                             /* the vector with the minimal dist. */
  r  = (int)((nbhood == SOM_RECT) ? radius : 3*radius);
  r2 = r*r;                     /* compute radius of region to adapt */
  xmin = bx -r; xmax = bx +r;   /* and with it compute the ranges of */
  ymin = by -r; ymax = by +r;   /* units in the two dimensions */
  if (xmin <  0)           xmin = 0;
  if (xmax >= som->width)  xmax = som->width -1;
  if (ymin <  0)           ymin = 0;
  if (ymax >= som->height) ymax = som->height-1;
  radius *= 2*radius;           /* precomp. for Gaussian function */
  for (iy = ymin; iy <= ymax; iy++) {
    rv = som->rvecs[iy] +xmin;  /* traverse the reference vectors */
    for (ix = xmin; ix <= xmax; ix++, rv++) {
      dxi = ix-bx; dyi = iy-by; /* compute the distance */
      di  = dxi*dxi + dyi*dyi;  /* to the winner neuron */
      if (di > r2) continue;    /* skip neurons that are too far away */
      if (nbhood == SOM_RECT) { /* if rectangular neighborhood */
        rv->x += lrate *(x -rv->x);
        rv->y += lrate *(y -rv->y);
        rv->z += lrate *(z -rv->z); }
      else {                    /* if Gaussian neighborhood */
        d = exp(-di/radius);
        rv->x += lrate *(x -rv->x) *d;
        rv->y += lrate *(y -rv->y) *d;
        rv->z += lrate *(z -rv->z) *d;
      }                         /* adapt the vectors, so that they */
    }                           /* get closer to the training pattern */
  }
}  /* som_train() */

/*--------------------------------------------------------------------*/

int som_save (SOM *som, FILE *file)
{                               /* --- save self-org. map to a file */
  int    x, y;                  /* loop variables */
  REFVEC *rv;                   /* to traverse the reference vectors */

  assert(som && file);          /* check the function arguments */
  fprintf(file, "%d %d\n", som->width, som->height);
  for (y = som->height; --y >= 0; ) {
    fprintf(file, "\n");        /* leave one line empty */
    rv = som->rvecs[y];         /* traverse the rows of the map */
    for (x = som->width; --x >= 0; rv++)
      fprintf(file, "%9.6f %9.6f %9.6f\n", rv->x, rv->y, rv->z);
  }                             /* print coordinates of the vectors */
  return (ferror(file)) ? -1 : 0;
}  /* som_save() */

/*--------------------------------------------------------------------*/

SOM* som_load (FILE *file)
{                               /* --- load self-org. map from a file */
  int    i;                     /* loop variable */
  int    width, height;         /* width and height of map */
  SOM    *som;                  /* created self-organizing map */
  REFVEC *rv;                   /* traverse the reference vectors */

  assert(file);                 /* check the function argument */
  if (fscanf(file, "%d%d", &width, &height) != 2)
    return NULL;                /* read width and height */
  som = som_create(width, height);
  if (!som) return NULL;        /* create a self-organizing map */
  while (--height >= 0) {       /* traverse the reference vectors */
    for (rv = som->rvecs[height], i = width; --i >= 0; rv++)
      if (fscanf(file, "%lf%lf%lf", &rv->x, &rv->y, &rv->z) != 3) {
        som_delete(som); return NULL; }
  }                             /* load the vector coordinates */
  return som;                   /* return the loaded self-org. map */
}  /* som_load() */

/*--------------------------------------------------------------------*/

void som_genpat (int shape, double randfn(void),
                 double *x, double *y, double *z)
{                               /* --- generate a training pattern */
  double tx, ty, tz;            /* generated coodinates */
  double norm;                  /* normalization factor */

  assert(randfn && x && y);     /* check the function arguments */
  tx = 2*randfn() -1;           /* generate two coordinates */
  ty = 2*randfn() -1;           /* uniformly distributed in [-1,1] */
  switch (shape) {              /* generate a training pattern */
    case SOM_TRIANGLE:          /* --- triangle */
      if      (ty >  2*tx+1) { tx += 1; ty = -ty; }
      else if (ty > -2*tx+1) { tx -= 1; ty = -ty; }
      *x = tx; *y = ty; *z = 0; /* fit coordinates to a triangle */
      break;                    /* (rotate certain regions) */

    case SOM_CIRCLE:            /* --- circle */
      while (tx*tx +ty*ty > 1) {
        tx = 2*randfn() -1;     /* while the point is outside */
        ty = 2*randfn() -1;     /* the unit circle, generate */
      }                         /* two new coordinates in [-1,1] */
      *x = tx; *y = ty; *z = 0; /* store the generated coordinates */
      break;                    /* and set the third coord. to zero */

    case SOM_SPHERE:            /* --- sphere surface */
      tz = 2*randfn() -1;       /* generate a third coordinate */
      norm = tx*tx + ty*ty + tz*tz;
      if (norm > 0) norm = 1/sqrt(norm);
      else     tx = norm = 1;
      *x = tx *norm; *y = ty *norm; *z = tz *norm;
      break;                    /* normalize the vector to length 1 */

    case SOM_PARABOLA:          /* --- rotated parabola */
      *x = tx; *y = ty;         /* store the generated coordinates */
      *z = tx*tx +ty*ty -1;     /* compute the third coordinate */
      break;

    case SOM_CUBIC:             /* --- cubic function */
      *x = tx; *y = ty;         /* store the generated coordinates */
      tx *= 1.2; ty *= 1.2;
      *z = (tx*tx*tx +ty*ty*ty -tx -ty);
      break;                    /* compute the third coordinate */

    case SOM_COSINE:            /* --- cosine function */
      *x = tx; *y = ty;         /* store the generated coordinates */
      norm = sqrt(tx*tx +ty*ty);
      *z   = 0.5*cos(2*M_PI *norm) /(norm+1);
      break;                    /* compute the third coordinate */

    case SOM_CUBE:              /* --- cube */
      *x = tx; *y = ty;         /* store the generated coordinates */
      *z = 2*randfn() -1;       /* and generate a third */
      break;

    default:                    /* --- square */
      *x = tx; *y = ty; *z = 0; /* store the generated coordinates */
      break;                    /* and set the third coord. to zero */
  }
}  /* som_genpat() */

/*--------------------------------------------------------------------*/
#ifdef SOM_DRAW

void som_draw (SOM *som, PROJ3D *p3d,
               void drawfn (double x, double y, int line))
{                               /* --- draw a self-organizing map */
  int    ix, iy;                /* loop variables, indices */
  REFVEC *rv;                   /* to traverse the reference vectors */

  rv = som->rvecs[0];           /* traverse the reference vectors */
  if (p3d) {                    /* if three-dimensional drawing */
    for (iy = som->height *som->width; --iy >= 0; rv++) {
      rv->proj = p3d_proj(p3d, rv->x, rv->y, rv->z, &rv->px, &rv->py);
    } }                         /* project all reference vectors */
  else {                        /* if two-dimensional drawing */
    for (iy = som->height *som->width; --iy >= 0; rv++) {
      rv->proj = 0; rv->px = rv->x; rv->py = rv->y; }
  }                             /* copy the first two coordinates */
  for (iy = som->height; --iy >= 0; ) {
    for (ix = 0; ix < som->width; ix++) {
      rv = som->rvecs[iy] +ix;  /* traverse the grid rows */
      drawfn(rv->px, rv->py, (rv->proj >= 0) ? ix : 0);
    }                           /* draw the grid rows */
  }
  for (ix = som->width; --ix >= 0; ) {
    for (iy = 0; iy < som->height; iy++) {
      rv = som->rvecs[iy] +ix;  /* traverse the grid columns */
      drawfn(rv->px, rv->py, (rv->proj >= 0) ? iy : 0);
    }                           /* draw the grid columns */
  }
}  /* som_draw() */

#endif
/*--------------------------------------------------------------------*/
#ifdef SOM_MPOST

void som_mpost (SOM *som, PROJ3D *p3d, FILE *file)
{                               /* --- draw map with METAPOST */
  int    ix, iy;                /* loop variables, indices */
  REFVEC *rv;                   /* to traverse the reference vectors */

  fprintf(file, "u  := 1mm;\npw := 1pt;\n\n");
  fprintf(file, "transform t; "
                  "t := identity scaled 16u shifted (17u,17u);\n\n");
  fprintf(file, "def refmark = fill (0,0)--(1,0)--(0,1)--cycle "
                  "withcolor 0.999white\nenddef;\n\n");
  fprintf(file, "vardef drawsom (expr w, h) =\n"
                "  refmark; pickup pencircle scaled pw;\n"
                "  for y = 0 upto h-1:\n"
                "    draw (p[y][0] for x = 1 upto w-1: "
                       "--p[y][x] endfor) transformed t;\n"
                "  endfor;\n"
                "  for x = 0 upto w-1:\n"
                "    draw (p[0][x] for y = 1 upto h-1: "
                       "--p[y][x] endfor) transformed t;\n"
                "  endfor;\n"
                "enddef;\n\n");
  fprintf(file, "beginfig(-1);\npair p[][];\n");
  rv = som->rvecs[0];           /* traverse the reference vectors */
  if (p3d) {                    /* if three-dimensional drawing */
    for (iy = som->height *som->width; --iy >= 0; rv++) {
      rv->proj = p3d_proj(p3d, rv->x, rv->y, rv->z, &rv->px, &rv->py);
    } }                         /* project all reference vectors */
  else {                        /* if two-dimensional drawing */
    for (iy = som->height *som->width; --iy >= 0; rv++) {
      rv->proj = 0; rv->px = rv->x; rv->py = rv->y; }
  }                             /* copy the first two coordinates */
  for (iy = som->height; --iy >= 0; ) {
    for (ix = 0; ix < som->width; ix++) {
      rv = som->rvecs[iy] +ix;
      fprintf(file, "p[%2d][%2d] = ", iy, ix);
      fprintf(file, "(% .3f,% .3f);\n", rv->px, rv->py);
    }                           /* draw one grid row */
  }
  fprintf(file, "drawsom(%d,%d);\n", som->width, som->height);
  fprintf(file, "endfig;\n\nend.\n");
}  /* som_mpost() */

#endif
